define([
    'require',
    'lodash',
    'plotly.js/dist/plotly.min',
    '@vienna/chart-lib'
], function (require, _, Plotly, charts) {
    'use strict';

    window.onresize = function () {
        try {
            resize();
        } catch (error) {
            console.info(error);
        }
    };

    function resize() {
        var d3 = Plotly.d3;
        var gd3 = $('.chart')
            .each(function (i, d) {
                // bypass hidden area which resize can't work
                if ($(d).is(":hidden")) {
                    return true;
                }
                var gd = d3.select(d).node();
                if (gd && gd.data) {
                    Plotly.Plots.resize(gd)
                };
            });
    }

    function _param(para, metricName) {
        return para === metricName ? 'best_metric' : ('param_' + para)
    }

    function _paramLabel(para, metricName) {
        return (para === metricName ? 'metric- ' : 'param- ') + para;
    }

    var chartModule = {
        // Function to render chart
        lineChart: function PlotLineChart(elementRef, chartData, showLegend, dTick, xtype, sortBy) {

            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };

            var top5;
            if (showLegend && sortBy) {
                chartData.series.sort((x, y) => {
                    return parseInt(x.run_id) - parseInt(y.run_id);
                });

                top5 = chartData.series.slice(0).sort((x, y) => {
                    return sortBy.order === 'asc' ?
                        (Math.min.apply(Math, y.data) - Math.min.apply(Math, x.data)) : (Math.max.apply(Math, y.data) - Math.max.apply(Math, x.data));
                }).map(r => r.run_id).slice(0, 5);
            }
            var metricData = chartData.series;
            // Rename data to y and add type as preferred by Plotly
            for (var i = 0; i < chartData.series.length; i++) {
                if (showLegend && top5 && top5.indexOf(parseInt(chartData.series[i].run_id)) === -1) {
                    metricData[i]['visible'] = 'legendonly';
                }
                metricData[i]['y'] = chartData.series[i]['data'];
                metricData[i]['x'] = chartData.series[i].categories ? chartData.series[i].categories : chartData.categories;
                if (chartData.series[i]['stepped'])
                    metricData[i]['line'] = {
                        shape: 'hv'
                    };
                delete chartData.series[i]['data'];
            };

            var chartMargin = {
                l: 40,
                r: 0,
                b: 40,
                t: 40,
                pad: 4
            }

            var layout = {
                autosize: true,
                title: chartData.title,
                margin: chartMargin,
                showlegend: showLegend,
                xaxis: {}
            }

            if (dTick) {
                layout.xaxis.dtick = dTick;
            }

            if (xtype) {
                layout.xaxis.type = xtype
            }

            Plotly.react(elementRef, metricData, layout, config);
        },
        precisionRecallChartProperties: function preparePRChartProperties(elementRef, chartMetrics, classLabels) {
            var overallRocTraceData = _.map(["macro", "micro"], function (label) {
                return {
                    xaxis_key: 'recall_' + label,
                    yaxis_key: 'precision_' + label,
                    trace_name: _.capitalize(label)
                };
            })

            var prChartProperties = {};
            prChartProperties.traceData = overallRocTraceData.concat(_.map(classLabels, function (label) {
                return {
                    xaxis_key: 'recall_class_' + label,
                    yaxis_key: 'precision_class_' + label,
                    trace_name: 'Class ' + label
                };
            }));
            prChartProperties.title = 'Multi-class Precision-Recall';
            prChartProperties.idealData = {
                x: [0, 1, 1],
                y: [1, 1, 0]
            }
            prChartProperties.xaxis_label = 'Recall';
            prChartProperties.yaxis_label = 'Precision';
            this.binaryClassificationScatterChart(elementRef, chartMetrics, prChartProperties);
        },
        rocChartProperties: function prepareRocChartProperties(elementRef, chartMetrics, classLabels) {
            var overallRocTraceData = _.map(["macro", "micro"], function (label) {
                return {
                    xaxis_key: 'false_positive_rate_' + label,
                    yaxis_key: 'true_positive_rate_' + label,
                    trace_name: _.capitalize(label)
                };
            })

            var rocChartProperties = {};
            rocChartProperties.traceData = overallRocTraceData.concat(_.map(classLabels, function (label) {
                return {
                    xaxis_key: 'false_positive_rate_class_' + label,
                    yaxis_key: 'true_positive_rate_class_' + label,
                    trace_name: 'Class ' + label
                };
            }));
            rocChartProperties.title = "ROC";
            rocChartProperties.idealData = {
                x: [0, 0, 1],
                y: [0, 1, 1]
            }
            rocChartProperties.xaxis_label = 'False Positive Rate';
            rocChartProperties.yaxis_label = 'True Positive Rate';
            this.binaryClassificationScatterChart(elementRef, chartMetrics, rocChartProperties);
        },

        binaryClassificationScatterChart: function plotBinaryClassificiationScatterChart(elementRef, metricData, chartProperties) {
            var chartData = _.map(chartProperties.traceData, function (t_data) {
                var trace = _.reduce(_.get(metricData, t_data.xaxis_key), function (result, xVal, idx) {
                    var yVal = _.get(metricData, [t_data.yaxis_key, idx]);
                    if (xVal === 'NaN' || yVal === 'NaN') {
                        return result;
                    }
                    result.x.push(xVal);
                    result.y.push(yVal);
                    return result;
                }, {
                    x: [],
                    y: []
                });
                trace.mode = 'lines+markers';
                trace.name = t_data.trace_name;
                if (trace.name !== 'Macro' && trace.name !== 'Micro') {
                    trace.mode = 'lines';
                    trace.visible = 'legendonly';
                    trace.line = {
                        dash: 'dot',
                        width: 4
                    };
                }
                return trace;
            });
            var idealTrace = {
                x: _.get(chartProperties, ['idealData', 'x']),
                y: _.get(chartProperties, ['idealData', 'y']),
                mode: 'lines',
                name: 'Ideal',
                line: {
                    dash: 'dashdot',
                    width: 4,
                    color: "rgb(49, 150, 63)"
                }
            };
            chartData.unshift(idealTrace);

            var layout = {
                title: chartProperties.title,
                xaxis: {
                    title: chartProperties.xaxis_label,
                    showgrid: true,
                    zeroline: true
                },
                yaxis: {
                    title: chartProperties.yaxis_label,
                    showline: false
                },
                hovermode: 'closest',
                autosize: true
            };
            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };
            Plotly.react(elementRef, chartData, layout, config);
        },

        confusionMatrix: function PlotConfusionMatrix(elementRef, confusion_matrices, labels) {
            var chart = new charts.ConfusionMatrixChart();
            var props = chart.BuildPlotlyProps(confusion_matrices, labels);
            Plotly.react(elementRef, props.Data, props.Layout, props.Config);
        },

        barChart: function PlotBarChart(elementRef, chartData, mode) {
            var config = {
                displayModeBar: false,
                displaylogo: false
            };

            var data = chartData;

            var chartMargin = {
                l: 40,
                r: 0,
                b: 30,
                t: 40,
                pad: 4
            }
            Plotly.react(elementRef, data, {
                autosize: true,
                margin: chartMargin,
                barmode: mode,
                showlegend: false,
                yaxis: {
                    showticklabels: false,
                    showgrid: false,
                }
            }, config);
        },
        horizontalBarChart: function PlotHorizontalBarChart(elementRef, list, mode) {
            var config = {
                displayModeBar: false,
                displaylogo: false
            };

            var chartData = []
            for (var i = 0; i < Object.keys(list).length; i++) {
                var key = Object.keys(list)[i];
                var val = Object.values(list)[i].count;
                chartData[i] = {
                    x: [val],
                    y: ['Status'],
                    name: key,
                    orientation: 'h',
                    type: 'bar',
                    text: key + ' (' + val + ')',
                    textposition: 'inside',
                    hoverinfo: 'none',
                    marker: {
                        color: Object.values(list)[i].color,
                        width: 1
                    },
                }
            }

            var chartMargin = {
                l: 0,
                r: 0,
                b: 0,
                t: 0,
                pad: 4
            }

            var layout = {
                autosize: true,
                margin: chartMargin,
                barmode: mode,
                showlegend: false,
                height: 50,
                xaxis: {
                    zeroline: false,
                    showticklabels: false,
                    showgrid: false,
                    tickangle: -45
                }
            };
            Plotly.react(elementRef, chartData, layout, config);
        },
        heatmapBarChart: function PlotHeatmapBarChart(elementRef, chartData) {
            var config = {
                displayModeBar: false,
                displaylogo: false
            };

            var gray = 20;
            var red = 50;
            var blue = 100;
            var scaleMap = {
                CompletedWorst: 0,
                CompletedBest: 0.1,
                NotStarted: 0.2,
                Failed: 0.5,
                Running: 1
            }
            var colorScaleValue = [
                [scaleMap.CompletedWorst, chartData.colors.lightgreen],
                [scaleMap.CompletedBest, chartData.colors.green],
                [scaleMap.NotStarted, chartData.colors.gray],
                [scaleMap.Failed, chartData.colors.red],
                [scaleMap.Running, chartData.colors.blue]
            ];

            var runs = _.sortBy(chartData.child_runs, 'run_number');
            var metrics = _.get(chartData, 'child_runs_metrics');
            var isMax = _.endsWith(_.get(metrics, 'series[1].name'), '_max');
            var categories = _.get(metrics, 'series[0].categories', []);
            var primaryMetrics = _.get(metrics, 'series[0].data', []);
            var metrics_mapping = _.zipObject(categories, primaryMetrics);
            var max = _.max(primaryMetrics);
            var min = _.min(primaryMetrics);

            var data = _.reduce(runs, function (result, run) {
                var iteration = run.iteration;
                var status = run.status;
                var metric = _.get(metrics_mapping, iteration);
                var color = gray;
                var text = iteration + ', ' + status;

                if (status === 'NotStarted') {
                    color = gray;
                } else if (status === 'Failed' || (status === 'Completed' && (_.isNaN(metric) || metric === 'NaN'))) {
                    color = red;
                } else if (status === 'Completed') {
                    if (max === min) {
                        color = 10;
                    } else if (isMax) {
                        color = 10 * (metric - min) / (max - min);
                    } else {
                        color = 10 * (max - metric) / (max - min);
                    }
                } else {
                    color = blue;
                }

                result.texts[0].push(text);
                result.xValues.push(iteration);
                result.zValues[0].push(color);

                return result;
            }, {
                texts: [
                    []
                ],
                xValues: [],
                yValues: ['Status'],
                zValues: [
                    []
                ]
            });

            var hm_data = {
                x: data.xValues,
                y: data.yValues,
                z: data.zValues,
                text: data.texts,
                hoverinfo: 'text',
                zmin: 0,
                zmax: 100,
                xgap: 3,
                ygap: 3,
                type: 'heatmap',
                colorscale: colorScaleValue,
                showscale: false,
                zauto: false,
                zsmooth: false,
                autocolorscale: false
            };

            var chartMargin = {
                l: 50,
                r: 0,
                b: 50,
                t: 10,
                pad: 4
            }

            var dtick = runs.length <= 10 ? 1 : Math.floor(runs.length / 10)
            if (dtick > 10) {
                dtick = _.floor(dtick, -1); // round down to 10
            } else if (dtick > 5) {
                dtick = 10;
            } else if (dtick > 2) {
                dtick = 5;
            }

            var layout = {
                xaxis: {
                    fixedrange: false,
                    autorange: true,
                    showgrid: false,
                    zeroline: false,
                    showticklabels: true,
                    type: 'category',
                    dtick: dtick
                },
                height: 100,
                margin: chartMargin,
            };

            Plotly.react(elementRef, [hm_data], layout, config);
        },
        changeableLineChart: function PlotChangeableLineChart(elementRef, dropElementRef, chartData, showLegend) {
            function getMetricData(chosenMetric) {
                var currentMetric;
                _.forOwn(chartData.series, function (metric) {
                    if (metric[0].name === chosenMetric) {
                        currentMetric = metric;
                    }
                })
                if (currentMetric) {
                    _.remove(currentMetric[0].data, function (metric) {
                        return metric === "NaN";
                    });
                    if (currentMetric[0].data.length < 1) {
                        return;
                    }
                }
                return currentMetric;
            }

            //Configurations for plotting chart
            function setBubblePlot(chosenMetric) {
                var metric = getMetricData(chosenMetric);
                if (_.isEmpty(metric)) {
                    return;
                }
                var trace1 = {
                    x: metric[0].categories,
                    y: metric[0].data,
                    mode: metric[0].mode,
                    marker: {
                        size: 12,
                        opacity: 0.8
                    },
                    name: metric[0].name
                };
                var trace2 = {
                    x: metric[1].categories,
                    y: metric[1].data,
                    mode: metric[1].mode,
                    line: {
                        width: 2,
                        shape: 'hv'
                    },
                    name: metric[1].name
                };

                var data = [trace1, trace2];

                var chartMargin = {
                    l: 40,
                    r: 0,
                    b: 40,
                    t: 80,
                    pad: 4
                }

                var layout = {
                    autosize: true,
                    title: 'AutoML Run with metric : ' + chosenMetric,
                    margin: chartMargin,
                    showlegend: showLegend,
                    type: 'category'
                }

                var config = {
                    modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                    displaylogo: false
                };

                Plotly.react(elementRef, data, layout, config);
            };

            function createOrUpdateDropdown(primaryMetricName) {
                var $sel = $(dropElementRef).find('#pm_dropdown');
                if ($sel.length < 1) {
                    $sel = $('<select id="pm_dropdown"></select>');
                }

                $sel.find('option').remove().end();
                _.forOwn(chartData.series, function (metric) {
                    var metric_obj = _.first(metric);
                    if (_.isEmpty(metric_obj) || !metric_obj.name || !_.isNumber(_.first(metric_obj.data))) {
                        return;
                    }

                    $('<option>').text(metric_obj.name).appendTo($sel);
                });
                if ($sel.has('option').length === 0) {
                    if ($(dropElementRef).find('#noMetricsContainer').length < 1) {
                        $(dropElementRef).append($('<div id=noMetricsContainer style="text-align:center">No metrics available.</div>'));
                    }
                    return;
                }
                $(dropElementRef).find('#noMetricsContainer').remove();
                dropElementRef.append($sel)
                $sel.val(primaryMetricName);
                $sel.off('change').on('change', function () {
                    setBubblePlot($(this).val());
                });
            }

            // Default plot to primary metric
            createOrUpdateDropdown(chartData.primaryMetricName);
            setBubblePlot(chartData.primaryMetricName);

        },
        predictedVsTrueChart: function PredictedVsTrueChart(elementRef, predictionsMetrics) {
            var chart = new charts.PredictedVsTrueChart();
            var props = chart.BuildPlotlyProps(predictionsMetrics)
            Plotly.react(elementRef, props.Data, props.Layout, props.Config);
        },
        /**
         * Draws a scatter plot in 2 dimensions.
         * `allArguments` accepts pairs of `[metricName, plotlyAxisType]` or just `metricName`.
         * 
         * @typedef {[string, ("linear"|"log"|"category"|"date"|"-")]} ChartAxisSpec
         * @typedef {(string|ChartAxisSpec)[]} ChartArgsList
         * @typedef {string[]} ChartChildRuns
         * @param {*} elementRef 
         * @param {({ metricName: string, allArguments: ChartArgsList, child_runs: ChartChildRuns })} chartData 
         * @param {*} heatMapOrder 
         */
        twoDimensionSurfaceChart: function PlotTwoDimensionSurfaceChart(elementRef, chartData, heatMapOrder) {
            if (!chartData || !chartData.child_runs || chartData.allArguments.length === 0) return;

            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };

            const child_runs = chartData.child_runs;
            const primaryMetricName = chartData.metricName;
            /**
             * Normalize the arguments into axis specs
             * @type {ChartAxisSpec[]}
             */
            const axisSpecs = chartData.allArguments.map(function parseChartDataArgument(arg) {
                return (typeof(arg) === "string") ? [arg, "-"] : arg;
            });
            axisSpecs.push([chartData.metricName, "-"]);
            const axisNames = axisSpecs.map(function (pair) { return pair[0]; });
            // Build a lookup of axis scalings by name
            const plotlyAxisTypesByName = axisSpecs.reduce(function(dict, spec) {
                dict[spec[0]] = spec[1];
                return dict;
            }, {});
            const initialXAxis = axisNames[0];
            const initialYAxis = axisNames.length > 1 ? axisNames[1] : axisNames[0];
            function getDataForParam(paramName) {
                return child_runs.map(data => data[_param(paramName, primaryMetricName)]);
            }
            function getAxisTypeForParam(paramName) {
                return plotlyAxisTypesByName[paramName];
            }
            function createAxisUpdate(paramName, axis) {
                if (axis != "x" && axis != "y") {
                    throw new Error(`Invalid axis fed into createAxisUpdate (2d): ${axis}`);
                }
                return {
                    method: 'update',
                    args: [
                        {
                            [axis]: [getDataForParam(paramName)]
                        },
                        {
                            [`${axis}axis.title`]: paramName,
                            [`${axis}axis.type`]: getAxisTypeForParam(paramName),
                        }
                    ],
                    label: _paramLabel(paramName, primaryMetricName)
                }
            }
            function sortAxesForButton(axisNames, primaryForAxis) {
                return axisNames.sort((a, b) => {
                    return a == primaryForAxis ? -1 : b == primaryForAxis ? 1 : 0;
                });
            }

            const data = [{
                x: getDataForParam(initialXAxis),
                y: getDataForParam(initialYAxis),
                text: child_runs.map(data => axisNames.map(para => para + ': ' + data[_param(para, primaryMetricName)]).join('<br>')),
                hoverinfo: 'text',
                mode: 'markers',
                marker: {
                    size: 6,
                    color: child_runs.map(data => data.best_metric),
                    colorscale: 'Viridis',
                    opacity: 0.8,
                    showscale: true,
                    reversescale: heatMapOrder === 'desc' ? false : (heatMapOrder === 'asc' ? true : undefined),
                    colorbar: {
                        len: 0.6,
                        x: 1,
                        xanchor: 'left',
                        yanchor: 'bottom',
                        y: 0.05,
                        ypad: 5,
                        title: 'Metric - ' + primaryMetricName,
                    }
                },
                type: 'scatter'
            }];

            const chartMargin = {
                l: 40,
                r: 0,
                b: 40,
                t: 40,
                pad: 4,
            };

            const layout = {
                autosize: true,
                margin: chartMargin,
                xaxis: {
                    title: initialXAxis,
                    type: getAxisTypeForParam(initialXAxis),
                },
                yaxis: {
                    title: initialYAxis,
                    type: getAxisTypeForParam(initialYAxis),
                },
                updatemenus: [{
                    x: 1,
                    y: 0.95,
                    xanchor: 'left',
                    yanchor: 'top',
                    showactive: true,
                    active: 0,
                    buttons: sortAxesForButton(axisNames, initialXAxis).map(newPara => createAxisUpdate(newPara, "x"))
                }, {
                    x: 1,
                    y: 0.85,
                    xanchor: 'left',
                    yanchor: 'top',
                    buttons: sortAxesForButton(axisNames, initialYAxis).map(newPara => createAxisUpdate(newPara, "y"))
                }],
            };

            Plotly.react(elementRef, data, layout, config);
        },
        /**
         * Draws a scatter plot in 3 dimensions.
         * `allArguments` accepts pairs of `[metricName, plotlyAxisType]` or just `metricName`.
         * 
         * @typedef {[string, ("linear"|"log"|"category"|"date"|"-")]} ChartAxisSpec
         * @typedef {(string|ChartAxisSpec)[]} ChartArgsList
         * @typedef {string[]} ChartChildRuns
         * @param {*} elementRef 
         * @param {({ metricName: string, allArguments: ChartArgsList, child_runs: ChartChildRuns })} chartData 
         * @param {*} heatMapOrder 
         */
        threeDimensionSurfaceChart: function PlotThreeDimensionSurfaceChart(elementRef, chartData, heatMapOrder) {
            if (!chartData || !chartData.child_runs || chartData.allArguments.length === 0) return;

            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };

            const child_runs = chartData.child_runs;
            const primaryMetricName = chartData.metricName;
            /**
             * Normalize the arguments into axis specs
             * @type {ChartAxisSpec[]}
             */
            const axisSpecs = chartData.allArguments.map(function parseChartDataArgument(arg) {
                return (typeof(arg) === "string") ? [arg, "-"] : arg;
            });
            axisSpecs.push([chartData.metricName, "-"]);
            const axisNames = axisSpecs.map(function (pair) { return pair[0]; });
            // Build a lookup of axis scalings by name
            const plotlyAxisTypesByName = axisSpecs.reduce(function(dict, spec) {
                dict[spec[0]] = spec[1];
                return dict;
            }, {});
            const initialXAxis = axisNames[0];
            const initialYAxis = axisNames.length > 1 ? axisNames[1] : axisNames[0];
            const initialZAxis = axisNames.length > 2 ? axisNames[2] : primaryMetricName;
            function getDataForParam(paramName) {
                return child_runs.map(data => data[_param(paramName, primaryMetricName)]);
            }
            function getAxisTypeForParam(paramName) {
                return plotlyAxisTypesByName[paramName];
            }
            function createAxisUpdate(paramName, axis) {
                if (axis != "x" && axis != "y" && axis != "z") {
                    throw new Error(`Invalid axis fed into createAxisUpdate (3d): ${axis}`);
                }
                return {
                    method: 'update',
                    args: [
                        {
                            [axis]: [getDataForParam(paramName)]
                        },
                        {
                            [`scene.${axis}axis.title`]: paramName,
                            [`scene.${axis}axis.type`]: getAxisTypeForParam(paramName),
                        }
                    ],
                    label: _paramLabel(paramName, primaryMetricName)
                }
            }
            function sortAxesForButton(axisNames, primaryForAxis) {
                return axisNames.sort((a, b) => {
                    return a == primaryForAxis ? -1 : b == primaryForAxis ? 1 : 0;
                });
            }

            const data = [{
                x: getDataForParam(initialXAxis),
                y: getDataForParam(initialYAxis),
                z: getDataForParam(initialZAxis),
                text: child_runs.map(data => axisNames.map(para => para + ': ' + data[_param(para, primaryMetricName)]).join('<br>')),
                hoverinfo: 'text',
                mode: 'markers',
                marker: {
                    size: 6,
                    color: child_runs.map(data => data.best_metric),
                    colorscale: 'Viridis',
                    opacity: 0.8,
                    showscale: true,
                    reversescale: heatMapOrder === 'desc' ? false : (heatMapOrder === 'asc' ? true : undefined),
                    colorbar: {
                        len: 0.6,
                        x: 1,
                        xanchor: 'left',
                        yanchor: 'bottom',
                        y: 0.05,
                        ypad: 5,
                        title: 'Metric - ' + primaryMetricName,
                    }
                },
                type: 'scatter3d'
            }];

            const chartMargin = {
                l: 0,
                r: 0,
                b: 0,
                t: 0,
                pad: 0,
            };

            const layout = {
                autosize: true,
                margin: chartMargin,
                scene: {
                    aspectratio: {
                        x: 0.9,
                        y: 0.9,
                        z: 0.9
                    },
                    xaxis: {
                        title: initialXAxis,
                        type: getAxisTypeForParam(initialXAxis),
                    },
                    yaxis: {
                        title: initialYAxis,
                        type: getAxisTypeForParam(initialYAxis),
                    },
                    zaxis: {
                        title: initialZAxis,
                        type: getAxisTypeForParam(initialZAxis),
                    }
                },
                updatemenus: [{
                        x: 1,
                        y: 0.95,
                        xanchor: 'left',
                        yanchor: 'top',
                        showactive: true,
                        active: 0,
                        buttons: sortAxesForButton(axisNames, initialXAxis).map(newPara => createAxisUpdate(newPara, "x"))
                    }, {
                        x: 1,
                        y: 0.85,
                        xanchor: 'left',
                        yanchor: 'top',
                        buttons: sortAxesForButton(axisNames, initialYAxis).map(newPara => createAxisUpdate(newPara, "y"))
                    },
                    {
                        x: 1,
                        y: 0.75,
                        xanchor: 'left',
                        yanchor: 'top',
                        buttons: sortAxesForButton(axisNames, initialZAxis).map(newPara => createAxisUpdate(newPara, "z"))
                    }
                ],
            };

            Plotly.react(elementRef, data, layout, config);
        },
        parallelCoordinatesChart: function PlotParallelCoordinatesChart(elementRef, chartData, heatMapOrder) {

            var allArguments = chartData.allArguments;
            var child_runs = chartData.child_runs;
            var metricName = chartData.metricName;

            var config = {
                displayModeBar: false,
                displaylogo: false
            };

            var layout = {
                autosize: true
            };

            var trace = {
                type: 'parcoords',
                line: {
                    color: child_runs.map(data => data.best_metric),
                    colorscale: 'Viridis',
                    opacity: 0.8,
                    showscale: true,
                    reversescale: heatMapOrder === 'desc' ? false : (heatMapOrder === 'asc' ? true : undefined),
                    colorbar: {
                        title: chartData.metricName,
                    }
                },
                dimensions: []
            };

            allArguments.forEach(arg => {
                trace.dimensions.push({
                    label: arg,
                    values: child_runs.map(data => data[_param(arg, metricName)])
                });
            });


            var data = [trace]

            Plotly.react(elementRef, data, layout, config);
        },
        featureImportanceChart: function PlotFeatureImportance(elementRef, feature_importance_metrics) {
            var chart = new charts.FeatureImportanceChart();
            var props = chart.BuildPlotlyProps(feature_importance_metrics);
            Plotly.react(elementRef, props.Data, props.Layout, props.Config);
        },

        pieChart: function PlotPieChart(elementRef, chartData, displayConfig = null) {
            // Apply defaults
            displayConfig = Object.assign({
                /**
                 * Expects an object with keys matching each label
                 *
                 * If falsey, colors are determined automatically.
                 * @throws {InvalidArgumentException} if truthy and not all labels are represented
                 */
                colorMap: false,

                /**
                 * Color for labels, otherwise uses the Plot.ly default.
                 *
                 * @type {?string}
                 */
                labelColor: null,

                /**
                 * The position for each label
                 *
                 * @type {("inside"|"outside")}
                 */
                labelPosition: "inside",

                /**
                 * Whether or not a legend should be shown for the chart
                 *
                 * @type {boolean}
                 */
                showLegend: false,
            }, displayConfig || {});

            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };

            var values = [];
            var labels = [];
            var textposition = [];

            for (var key in chartData) {
                if (chartData.hasOwnProperty(key)) {
                    values.push(chartData[key]);
                    labels.push(key);
                    // Do not display labels for segments with values of zero
                    textposition.push(chartData[key] === 0 ? 'none' : displayConfig.labelPosition);
                }
            }

            var data = {
                values: values,
                labels: labels,
                marker: {},
                textinfo: 'label+value',
                textposition: textposition,
                type: 'pie'
            };

            if (displayConfig.colorMap) {
                const colors = [];
                const missingSectorColors = [];
                for (let label of labels) {
                    if (displayConfig.colorMap.hasOwnProperty(label)) {
                        colors.push(displayConfig.colorMap[label]);
                    } else {
                        missingSectorColors.push(label);
                        colors.push("gray");
                    }
                }
                if (missingSectorColors.length > 0) {
                    if (console && console.error) {
                        console.error(
                            "PlotPieChart - Colors not provided for labeled sections [" +
                            missingSectorColors.join(", ") +
                            "]");
                    }
                }
                data.marker.colors = colors;
            }

            if ((typeof (displayConfig.labelColor)) === "string") {
                data.textfont = {
                    color: displayConfig.labelColor,
                };
            }

            var chartMargin = {
                l: 0,
                r: 0,
                b: 20,
                t: 20,
                pad: 4
            }

            var layout = {
                height: 140,
                autosize: true,
                margin: chartMargin,
                showlegend: displayConfig.showLegend,
                xaxis: {}
            }

            Plotly.react(elementRef, [data], layout, config);
        },
        purge: function PlotPurge(elementRef) {
            Plotly.purge(elementRef);
        },

    }

    return chartModule;
});